@@ -28,11 +28,12 @@
  */
 
 #define	_GNU_SOURCE
+#define _FILE_OFFSET_BITS 64
 #include	<unistd.h>
-#ifndef __dietlibc__
+#if !defined(__dietlibc__) && !defined(__KLIBC__)
 extern __off64_t lseek64 __P ((int __fd, __off64_t __offset, int __whence));
 #else
-# ifdef __NO_STAT64
+# if defined(__NO_STAT64) || __WORDSIZE != 32
 # define lseek64 lseek
 # endif
 #endif
@@ -97,6 +98,7 @@ extern __off64_t lseek64 __P ((int __fd, __off64_t __offset, int __whence));
 		     ((x) & 0x00000000ff000000ULL) << 8 | \
 		     ((x) & 0x000000ff00000000ULL) >> 8)
 
+#if !defined(__KLIBC__)
 #if BYTE_ORDER == LITTLE_ENDIAN
 #define	__cpu_to_le16(_x) (_x)
 #define __cpu_to_le32(_x) (_x)
@@ -114,6 +116,7 @@ extern __off64_t lseek64 __P ((int __fd, __off64_t __offset, int __whence));
 #else
 #  error "unknown endianness."
 #endif
+#endif /* __KLIBC__ */
 
 
 
@@ -129,6 +132,14 @@ struct mdinfo {
 	int			reshape_active;
 	unsigned long long	reshape_progress;
 	int			new_level, delta_disks, new_layout, new_chunk;
+	int			errors;
+	int			cache_size; /* size of raid456 stripe cache*/
+	int			mismatch_cnt;
+	char			text_version[50];
+
+	char 		sys_name[20];
+	struct mdinfo *devs;
+	struct mdinfo *next;
 };
 
 struct createinfo {
@@ -151,9 +162,11 @@ enum mode {
 	MONITOR,
 	GROW,
 	INCREMENTAL,
+	AUTODETECT,
 };
 
 extern char short_options[];
+extern char short_bitmap_options[];
 extern char short_bitmap_auto_options[];
 extern struct option long_options[];
 extern char Version[], Usage[], Help[], OptionHelp[],
@@ -175,6 +188,7 @@ enum special_options {
 	HomeHost,
 	AutoHomeHost,
 	Symlinks,
+	AutoDetect,
 };
 
 /* structures read from config file */
@@ -190,7 +204,7 @@ enum special_options {
 #define UnSet (0xfffe)
 typedef struct mddev_ident_s {
 	char	*devname;
-	
+
 	int	uuid_set;
 	int	uuid[4];
 	char	name[33];
@@ -218,7 +232,7 @@ typedef struct mddev_dev_s {
 	char disposition;	/* 'a' for add, 'r' for remove, 'f' for fail.
 				 * Not set for names read from .config
 				 */
-	char writemostly;
+	char writemostly;	/* 1 for 'set writemostly', 2 for 'clear writemostly' */
 	char re_add;
 	char used;		/* set when used */
 	struct mddev_dev_s *next;
@@ -263,28 +277,6 @@ extern void map_free(struct map_ent *map);
 extern void map_add(struct map_ent **melp,
 		    int devnum, int major, int minor, int uuid[4], char *path);
 
-/* Data structure for holding info read from sysfs */
-struct sysdev {
-	char	name[20];
-	int	role;
-	int	major, minor;
-	unsigned long long offset, size;
-	int	state;
-	int	errors;
-	struct sysdev *next;
-};
-struct sysarray {
-	char	name[20];
-	struct sysdev *devs;
-	int	chunk;
-	unsigned long long component_size;
-	int	layout;
-	int	level;
-	int	spares;
-	int	cache_size;
-	int	mismatch_cnt;
-	int	major_version, minor_version;
-};
 /* various details can be requested */
 #define	GET_LEVEL	1
 #define	GET_LAYOUT	2
@@ -303,13 +295,13 @@ struct sysarray {
 /* If fd >= 0, get the array it is open on,
  * else use devnum. >=0 -> major9. <0.....
  */
-extern void sysfs_free(struct sysarray *sra);
-extern struct sysarray *sysfs_read(int fd, int devnum, unsigned long options);
-extern int sysfs_set_str(struct sysarray *sra, struct sysdev *dev,
+extern void sysfs_free(struct mdinfo *sra);
+extern struct mdinfo *sysfs_read(int fd, int devnum, unsigned long options);
+extern int sysfs_set_str(struct mdinfo *sra, struct mdinfo *dev,
 			 char *name, char *val);
-extern int sysfs_set_num(struct sysarray *sra, struct sysdev *dev,
+extern int sysfs_set_num(struct mdinfo *sra, struct mdinfo *dev,
 			 char *name, unsigned long long val);
-extern int sysfs_get_ll(struct sysarray *sra, struct sysdev *dev,
+extern int sysfs_get_ll(struct mdinfo *sra, struct mdinfo *dev,
 			char *name, unsigned long long *val);
 
 
@@ -336,29 +328,36 @@ extern char *map_dev(int major, int minor, int create);
 
 
 extern struct superswitch {
-	void (*examine_super)(void *sbv, char *homehost);
-	void (*brief_examine_super)(void *sbv);
-	void (*detail_super)(void *sbv, char *homehost);
-	void (*brief_detail_super)(void *sbv);
-	void (*uuid_from_super)(int uuid[4], void *sbv);
-	void (*getinfo_super)(struct mdinfo *info, void *sbv);
-	int (*match_home)(void *sbv, char *homehost);
-	int (*update_super)(struct mdinfo *info, void *sbv, char *update,
+	void (*examine_super)(struct supertype *st, char *homehost);
+	void (*brief_examine_super)(struct supertype *st);
+	void (*export_examine_super)(struct supertype *st);
+	void (*detail_super)(struct supertype *st, char *homehost);
+	void (*brief_detail_super)(struct supertype *st);
+	void (*export_detail_super)(struct supertype *st);
+	void (*uuid_from_super)(struct supertype *st, int uuid[4]);
+	void (*getinfo_super)(struct supertype *st, struct mdinfo *info);
+	int (*match_home)(struct supertype *st, char *homehost);
+	int (*update_super)(struct supertype *st, struct mdinfo *info,
+			    char *update,
 			    char *devname, int verbose,
 			    int uuid_set, char *homehost);
-	int (*init_super)(struct supertype *st, void **sbp, mdu_array_info_t *info, unsigned long long size, char *name, char *homehost, int *uuid);
-	void (*add_to_super)(void *sbv, mdu_disk_info_t *dinfo);
-	int (*store_super)(struct supertype *st, int fd, void *sbv);
-	int (*write_init_super)(struct supertype *st, void *sbv, mdu_disk_info_t *dinfo, char *devname);
-	int (*compare_super)(void **firstp, void *secondv);
-	int (*load_super)(struct supertype *st, int fd, void **sbp, char *devname);
+	int (*init_super)(struct supertype *st, mdu_array_info_t *info,
+			  unsigned long long size, char *name,
+			  char *homehost, int *uuid);
+	void (*add_to_super)(struct supertype *st, mdu_disk_info_t *dinfo);
+	int (*store_super)(struct supertype *st, int fd);
+	int (*write_init_super)(struct supertype *st, mdu_disk_info_t *dinfo,
+				char *devname);
+	int (*compare_super)(struct supertype *st, struct supertype *tst);
+	int (*load_super)(struct supertype *st, int fd, char *devname);
 	struct supertype * (*match_metadata_desc)(char *arg);
 	__u64 (*avail_size)(struct supertype *st, __u64 size);
-	int (*add_internal_bitmap)(struct supertype *st, void *sbv, int *chunkp,
+	int (*add_internal_bitmap)(struct supertype *st, int *chunkp,
 				   int delay, int write_behind,
 				   unsigned long long size, int may_change, int major);
-	void (*locate_bitmap)(struct supertype *st, int fd, void *sbv);
-	int (*write_bitmap)(struct supertype *st, int fd, void *sbv);
+	void (*locate_bitmap)(struct supertype *st, int fd);
+	int (*write_bitmap)(struct supertype *st, int fd);
+	void (*free_super)(struct supertype *st);
 	int major;
 	int swapuuid; /* true if uuid is bigending rather than hostendian */
 } super0, super1, *superlist[];
@@ -367,10 +366,12 @@ struct supertype {
 	struct superswitch *ss;
 	int minor_version;
 	int max_devs;
+	void *sb;
 };
 
-extern struct supertype *super_by_version(int vers, int minor);
+extern struct supertype *super_by_fd(int fd);
 extern struct supertype *guess_super(int fd);
+extern struct supertype *dup_super(struct supertype *st);
 extern int get_dev_size(int fd, char *dname, unsigned long long *sizep);
 extern void get_one_disk(int mdfd, mdu_array_info_t *ainf,
 			 mdu_disk_info_t *disk);
@@ -382,8 +383,11 @@ struct stat64;
 #define HAVE_NFTW  we assume
 #define HAVE_FTW
 
-#ifdef UCLIBC
+#ifdef __UCLIBC__
 # include <features.h>
+# ifndef __UCLIBC_HAS_LFS__
+#  define lseek64 lseek
+# endif
 # ifndef  __UCLIBC_HAS_FTW__
 #  undef HAVE_FTW
 #  undef HAVE_NFTW
@@ -394,6 +398,11 @@ struct stat64;
 # undef HAVE_NFTW
 #endif
 
+#if defined(__KLIBC__)
+# undef HAVE_NFTW
+# undef HAVE_FTW
+#endif
+
 #ifndef HAVE_NFTW
 # define FTW_PHYS 1
 # ifndef HAVE_FTW
@@ -414,6 +423,7 @@ extern int Manage_resize(char *devname, int fd, long long size, int raid_disks);
 extern int Manage_reconfig(char *devname, int fd, int layout);
 extern int Manage_subdevs(char *devname, int fd,
 			  mddev_dev_t devlist, int verbose);
+extern int autodetect(void);
 extern int Grow_Add_device(char *devname, int fd, char *newdev);
 extern int Grow_addbitmap(char *devname, int fd, char *file, int chunk, int delay, int write_behind, int force);
 extern int Grow_reshape(char *devname, int fd, int quiet, char *backup_file,
@@ -443,10 +453,10 @@ extern int Create(struct supertype *st, char *mddev, int mdfd,
 		  int runstop, int verbose, int force, int assume_clean,
 		  char *bitmap_file, int bitmap_chunk, int write_behind, int delay);
 
-extern int Detail(char *dev, int brief, int test, char *homehost);
+extern int Detail(char *dev, int brief, int export, int test, char *homehost);
 extern int Query(char *dev);
-extern int Examine(mddev_dev_t devlist, int brief, int scan, int SparcAdjust,
-		   struct supertype *forcest, char *homehost);
+extern int Examine(mddev_dev_t devlist, int brief, int export, int scan,
+		   int SparcAdjust, struct supertype *forcest, char *homehost);
 extern int Monitor(mddev_dev_t devlist,
 		   char *mailaddr, char *alert_cmd,
 		   int period, int daemonise, int scan, int oneshot,
@@ -467,6 +477,7 @@ extern int CreateBitmap(char *filename, int force, char uuid[16],
 			int major);
 extern int ExamineBitmap(char *filename, int brief, struct supertype *st);
 extern int bitmap_update_uuid(int fd, int *uuid, int swap);
+extern unsigned long bitmap_sectors(struct bitmap_super_s *bsb);
 
 extern int md_get_version(int fd);
 extern int get_linux_version(void);
@@ -496,7 +507,6 @@ extern int match_oneof(char *devices, char *devname);
 extern void uuid_from_super(int uuid[4], mdp_super_t *super);
 extern int same_uuid(int a[4], int b[4], int swapuuid);
 extern void copy_uuid(void *a, int b[4], int swapuuid);
-/* extern int compare_super(mdp_super_t *first, mdp_super_t *second);*/
 extern unsigned long calc_csum(void *super, int bytes);
 extern int enough(int level, int raid_disks, int layout, int clean,
 		   char *avail, int avail_disks);
@@ -506,7 +516,11 @@ extern void remove_partitions(int fd);
 
 
 extern char *human_size(long long bytes);
-char *human_size_brief(long long bytes);
+extern char *human_size_brief(long long bytes);
+extern void print_r10_layout(int layout);
+
+#define NoMdDev (1<<23)
+extern int find_free_devnum(int use_partitions);
 
 extern void put_md_name(char *name);
 extern char *get_md_name(int dev);
@@ -515,8 +529,19 @@ extern char DefaultConfFile[];
 
 extern int open_mddev(char *dev, int autof);
 extern int open_mddev_devnum(char *devname, int devnum, char *name,
-			     char *chosen_name);
+			     char *chosen_name, int parts);
 
+#include <assert.h>
+#include <stdarg.h>
+static inline int xasprintf(char **strp, const char *fmt, ...) {
+	va_list ap;
+	int ret;
+	va_start(ap, fmt);
+	ret = vasprintf(strp, fmt, ap);
+	va_end(ap);
+	assert(ret >= 0);
+	return ret;
+}
 
 #define	LEVEL_MULTIPATH		(-4)
 #define	LEVEL_LINEAR		(-1)
